/*	ppusbcomm : USB to Parallel port transfer
	Copyright (C) 2015  Raphael Assenat <raph@raphnet.net>

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "main.h"
#include "util.h"
#include "usart1.h"
#include "usb.h"
#include "bootloader.h"

#include "eeprom.h"
#include "hiddata.h"
#include "usbstrings.h"

/* Those .c files are included rather than linked for we
 * want the sizeof() operator to work on the arrays */
#include "dataHidReport.c"

#define MAX_READ_ERRORS	30

struct cfg0 {
	struct usb_configuration_descriptor configdesc;
	struct usb_interface_descriptor interface_admin;
	struct usb_hid_descriptor hid_data;
	struct usb_endpoint_descriptor ep1_in;
};

static const struct cfg0 cfg0 PROGMEM = {
	.configdesc = {
		.bLength = sizeof(struct usb_configuration_descriptor),
		.bDescriptorType = CONFIGURATION_DESCRIPTOR,
		.wTotalLength = sizeof(cfg0), // includes all descriptors returned together
		.bNumInterfaces = 1,
		.bConfigurationValue = 1,
		.bmAttributes = CFG_DESC_ATTR_RESERVED, // set Self-powred and remote-wakeup here if needed.
		.bMaxPower = 25, // for 50mA
	},

	// HID interface for config and update
	.interface_admin = {
		.bLength = sizeof(struct usb_interface_descriptor),
		.bDescriptorType = INTERFACE_DESCRIPTOR,
		.bInterfaceNumber = 0,
		.bAlternateSetting = 0,
		.bNumEndpoints = 1,
		.bInterfaceClass = USB_DEVICE_CLASS_HID,
		.bInterfaceSubClass = HID_SUBCLASS_NONE,
		.bInterfaceProtocol = HID_PROTOCOL_NONE,
	},
	.hid_data = {
		.bLength = sizeof(struct usb_hid_descriptor),
		.bDescriptorType = HID_DESCRIPTOR,
		.bcdHid = 0x0101,
		.bCountryCode = HID_COUNTRY_NOT_SUPPORTED,
		.bNumDescriptors = 1, // Only a report descriptor
		.bClassDescriptorType = REPORT_DESCRIPTOR,
		.wClassDescriptorLength = sizeof(dataHidReport),
	},
	.ep1_in = {
		.bLength = sizeof(struct usb_endpoint_descriptor),
		.bDescriptorType = ENDPOINT_DESCRIPTOR,
		.bEndpointAddress = USB_RQT_DEVICE_TO_HOST | 1, // 0x81
		.bmAttributes = TRANSFER_TYPE_INT,
		.wMaxPacketsize = 64,
		.bInterval = LS_FS_INTERVAL_MS(1),
	},

};

const struct usb_device_descriptor device_descriptor PROGMEM = {
	.bLength = sizeof(struct usb_device_descriptor),
	.bDescriptorType = DEVICE_DESCRIPTOR,
	.bcdUSB = 0x0101,
	.bDeviceClass = 0, // set at interface
	.bDeviceSubClass = 0, // set at interface
	.bDeviceProtocol = 0,
	.bMaxPacketSize = 64,
	.idVendor = 0x289B,
	.idProduct = 0x001C,
	.bcdDevice = 0x0100, // 1.0.0
	.bNumConfigurations = 1,
	.iManufacturer = 1,
	.iProduct = 2,
	.iSerialNumber = 3,
};

/** **/

static struct usb_parameters usb_params = {
	.flags = USB_PARAM_FLAG_CONFDESC_PROGMEM |
				USB_PARAM_FLAG_DEVDESC_PROGMEM |
					USB_PARAM_FLAG_REPORTDESC_PROGMEM,
	.devdesc = (PGM_VOID_P)&device_descriptor,
	.configdesc = (PGM_VOID_P)&cfg0,
	.configdesc_ttllen = sizeof(cfg0),
	.num_strings = NUM_USB_STRINGS,
	.strings = g_usb_strings,

	.n_hid_interfaces = 1,
	.hid_params = {
		[0] = {
			.reportdesc = dataHidReport,
			.reportdesc_len = sizeof(dataHidReport),
			.getReport = hiddata_get_report,
			.setReport = hiddata_set_report,
		},
	},
};

void hwinit(void)
{
	/* PORTB
	 *
	 * 7: NC    Input pull up
	 * 6: NC	Input pull up
	 * 5: NC	Input pull up
	 * 4: NC	Input pull up
	 * 3: MISO	Input pull up (parport DATA 2)
	 * 2: MOSI	Input pull up (parport DATA 1)
	 * 1: SCK	Input pull up (parport DATA 0)
	 * 0: NC	Input pull up
	 */
	DDRB = 0x00;
	PORTB = 0xFF;


	/* PORTC
	 *
	 * 7: NC	Output low
	 * 6: NC	Output low
	 * 5: NC	Output low
	 * 4: NC	Output low
	 * 3: (no such pin)
	 * 2: NC	Output low
	 * 1: RESET	(N/A: Reset input per fuses) (left "floating")
	 * 0: XTAL2	(N/A: Crystal oscillator) (left "floating")
	 */
	DDRC = 0xFC;
	PORTC = 0x00;

	/* PORTD
	 *
	 * 7: HWB		Input (external pull-up)
	 * 6: NC		Output low
	 * 5: NC		Output low
	 * 4: NC		Output low
	 *
	 * 3: IO3_MCU	Output low
	 * 2: IO1_MCU	Output low
	 * 1: IO2_MCU	Output low
	 * 0: IO0_MCU	Output low
	 */
	PORTD = 0x00;
	DDRD = 0x7F;

	// System clock. External crystal is 16 Mhz and we want
	// to run at max. speed. (no prescaler)
	CLKPR = 0x80;
	CLKPR = 0x0; // Division factor of 1
	PRR0 = 0;
	PRR1 = 0;
}

/* Called after eeprom content is loaded. */
void eeprom_app_ready(void)
{
	static wchar_t serial_from_eeprom[SERIAL_NUM_LEN+1];
	int i;

	for (i=0; i<SERIAL_NUM_LEN; i++) {
		serial_from_eeprom[i] = g_eeprom_data.cfg.serial[i];
	}
	serial_from_eeprom[i] = 0;
	g_usb_strings[USB_STRING_SERIAL_IDX] = serial_from_eeprom;
}

#define GET_IN()	((PINB & (0xE)) >> 1)

static inline void SET_OUT(unsigned char b)
{
	unsigned char val = 0;

	val = (b & 0x08) ^ 0x08; // Parport BUSY is inverted
	val |= (b & 0x04) >> 1;
	val |= (b & 0x02) << 1;
	val |= b & 0x01;

	PORTD = val;
}

volatile unsigned char g_block_valid = 0;
volatile unsigned char g_blockdata[DATA_BLOCK_SIZE];
static unsigned char block_nibble;

static void onCmdChanged(unsigned char cmd)
{
	switch (cmd)
	{
		case 1:
			SET_OUT(0);
			break;
		case 0:
			if (g_block_valid) {
				cli();
			} else {
				sei();
			}
			SET_OUT(g_block_valid); // block available
			block_nibble = 0;
			break;
		case 2:
		case 3:
			if (g_block_valid) {
				if (block_nibble & 1) {
					SET_OUT(g_blockdata[block_nibble>>1] & 0xf);
				} else {
					SET_OUT(g_blockdata[block_nibble>>1] >> 4);
				}
				block_nibble++;

				if (block_nibble >= DATA_BLOCK_SIZE*2) {
					g_block_valid = 0;
					break;
				}
			}
			break;
	}
}

int main(void)
{
	unsigned char cmd, prev_cmd = 0xff;
	int i;

	hwinit();
	usart1_init();
	eeprom_init();

	sei();
	usb_init(&usb_params);

	while (1)
	{
		usb_doTasks();
		hiddata_doTask();

		cmd = GET_IN();
		if (cmd != prev_cmd) {
			_delay_us(1);
			cmd = GET_IN();
			onCmdChanged(cmd);
		}

		prev_cmd = cmd;
	}

	return 0;
}
