/*	ppusbcomm : USB to Parallel port transfer
	Copyright (C) 2015  Raphael Assenat <raph@raphnet.net>

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>
#include <unistd.h>
#include <wchar.h>

#include "hexdump.h"
#include "version.h"
#include "raphnethid.h"
#include "ppusbcomm.h"
#include "../ppusbcomm_requests.h"

static void printUsage(void)
{
	printf("./ppusbcomm [OPTION]... [COMMAND]....\n");
	printf("Control tool for GC/N64 to USB adapter. Version %s\n", VERSION_STR);
	printf("\n");
	printf("Options:\n");
	printf("  -h, --help            Print help\n");
	printf("  -l, --list            List devices\n");
	printf("  -s serial             Operate on specified device (required unless -f is specified)\n");
	printf("  -f, --force           If no serial is specified, use first device detected.\n");
	printf("  -u, --upload file     Upload a file\n");
	//printf("  -i, --infile file     Input file for write operations (eg: --gc_to_n64_update)\n");
	printf("\n");
	printf("Configuration commands:\n");
	printf("  --get_version                      Read adapter firmware version\n");
	printf("  --set_serial serial                Assign a new device serial number\n");
	printf("  --get_serial                       Read serial from eeprom\n");
	printf("\n");
	printf("Advanced commands:\n");
	printf("  --bootloader                       Re-enumerate in bootloader mode\n");
	printf("  --noheader                         Do not send a transfer header block\n");
	printf("  --dry-run                          Do not actually upload the file\n");
}


#define OPT_SET_SERIAL				257
#define OPT_GET_SERIAL				258
#define OPT_UPLOAD					'u'
#define OPT_BOOTLOADER				259
#define OPT_GET_VERSION				260
#define OPT_GET_SIGNATURE			261
#define OPT_NOHEADER				262
#define OPT_DRY_RUN					263

struct option longopts[] = {
	{ "help", 0, NULL, 'h' },
	{ "list", 0, NULL, 'l' },
	{ "force", 0, NULL, 'f' },
	{ "serial", 0, NULL, 's' },
	{ "set_serial", 1, NULL, OPT_SET_SERIAL },
	{ "get_serial", 0, NULL, OPT_GET_SERIAL },
	{ "bootloader", 0, NULL, OPT_BOOTLOADER },
	{ "get_version", 0, NULL, OPT_GET_VERSION },
	{ "get_signature", 0, NULL, OPT_GET_SIGNATURE },
	{ "upload", 1, NULL, 'u' },
	{ "noheader", 0, NULL, OPT_NOHEADER },
	{ "dry-run", 0, NULL, OPT_DRY_RUN },
	{ },
};

static int listDevices(void)
{
	int n_found = 0;
	struct raphnethid_list_ctx *listctx;
	struct raphnethid_info inf;

	listctx = raphnethid_allocListCtx();
	if (!listctx) {
		fprintf(stderr, "List context could not be allocated\n");
		return -1;
	}
	while (raphnethid_listDevices(&inf, listctx))
	{
		n_found++;
		printf("Found device '%ls', serial '%ls'\n", inf.str_prodname, inf.str_serial);
	}
	raphnethid_freeListCtx(listctx);
	printf("%d device(s) found\n", n_found);

	return n_found;
}

int main(int argc, char **argv)
{
	raphnethid_hdl_t hdl;
	struct raphnethid_list_ctx *listctx;
	int opt, retval = 0;
	struct raphnethid_info inf;
	struct raphnethid_info *selected_device = NULL;
	int verbose = 0, use_first = 0, serial_specified = 0;
	int cmd_list = 0;
	unsigned int upload_flags = 0;
#define TARGET_SERIAL_CHARS 128
	wchar_t target_serial[TARGET_SERIAL_CHARS];
	const char *short_optstr = "hls:fu:";
	const char *outfile = NULL;
	const char *infile = NULL;

	while((opt = getopt_long(argc, argv, short_optstr, longopts, NULL)) != -1) {
		switch(opt)
		{
			case 's':
				{
					mbstate_t ps;
					memset(&ps, 0, sizeof(ps));
					if (mbsrtowcs(target_serial, (const char **)&optarg, TARGET_SERIAL_CHARS, &ps) < 1) {
						fprintf(stderr, "Invalid serial number specified\n");
						return -1;
					}
					serial_specified = 1;
				}
				break;
			case 'f':
				use_first = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'h':
				printUsage();
				return 0;
			case 'l':
				cmd_list = 1;
				break;
			case 'o':
				outfile = optarg;
				printf("Output file: %s\n", outfile);
				break;
			case 'u':
				infile = optarg;
				printf("Input file: %s\n", infile);
				break;
			case '?':
				fprintf(stderr, "Unrecognized argument. Try -h\n");
				return -1;
			case OPT_NOHEADER:
				upload_flags |= PPUSBCOMM_UPLOAD_FLAG_OMIT_HEADER;
				break;
			case OPT_DRY_RUN:
				upload_flags |= PPUSBCOMM_UPLOAD_FLAG_DRY_RUN;
				break;
		}
	}

	raphnethid_init(verbose);

	if (cmd_list) {
		printf("Simply listing the devices...\n");
		return listDevices();
	}

	if (!serial_specified && !use_first) {
		fprintf(stderr, "A serial number or -f must be used. Try -h for more information.\n");
		return 1;
	}

	listctx = raphnethid_allocListCtx();
	while ((selected_device = raphnethid_listDevices(&inf, listctx)))
	{
		if (serial_specified) {
			if (0 == wcscmp(inf.str_serial, target_serial)) {
				break;
			}
		}
		else {
			// use_first == 1
			printf("Will use device '%ls' serial '%ls'\n", inf.str_prodname, inf.str_serial);
			break;
		}
	}
	raphnethid_freeListCtx(listctx);

	if (!selected_device) {
		if (serial_specified) {
			fprintf(stderr, "Device not found\n");
		} else {
			fprintf(stderr, "No device found\n");
		}
		return 1;
	}

	hdl = raphnethid_openDevice(selected_device);
	if (!hdl) {
		printf("Error opening device. (Do you have permissions?)\n");
		return 1;
	}

	optind = 1;
	while((opt = getopt_long(argc, argv, short_optstr, longopts, NULL)) != -1)
	{
		unsigned char cmd[64] = { };
		int n;
		int cmdlen = 0;
		int do_exchange = 0;

		switch (opt)
		{
			case OPT_SET_SERIAL:
				printf("Setting serial...");
				if (strlen(optarg) != 6) {
					fprintf(stderr, "Serial number must be 6 characters\n");
					return -1;
				}
				ppusbcomm_setConfig(hdl, CFG_PARAM_SERIAL, (void*)optarg, 6);
				break;

			case OPT_GET_SERIAL:
				n = ppusbcomm_getConfig(hdl, CFG_PARAM_SERIAL, cmd, sizeof(cmd));
				if (n==6) {
					cmd[6] = 0;
					printf("Serial: %s\n", cmd);
				}
				break;

			case OPT_BOOTLOADER:
				printf("Sending 'jump to bootloader' command...");
				ppusbcomm_bootloader(hdl);
				break;

			case OPT_GET_VERSION:
				{
					char version[64];

					if (0 == ppusbcomm_getVersion(hdl, version, sizeof(version))) {
						printf("Firmware version: %s\n", version);
					}
				}
				break;

			case OPT_GET_SIGNATURE:
				{
					char sig[64];

					if (0 == ppusbcomm_getSignature(hdl, sig, sizeof(sig))) {
						printf("Signature: %s\n", sig);
					}
				}
				break;

			case OPT_UPLOAD:
				{
					ppusbcomm_uploadFile(hdl, infile, upload_flags);
				}
				break;
		}

		if (do_exchange) {
			int i;
			n = raphnethid_exchange(hdl, cmd, cmdlen, cmd, sizeof(cmd));
			if (n<0)
				break;

			printf("Result: %d bytes: ", n);
			for (i=0; i<n; i++) {
				printf("%02x ", cmd[i]);
			}
			printf("\n");
		}
	}

	raphnethid_closeDevice(hdl);
	raphnethid_shutdown();

	return retval;
}
