/*	ppusbcomm : USB to Parallel port transfer
	Copyright (C) 2015  Raphael Assenat <raph@raphnet.net>

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdio.h>
#include "ppusbcomm.h"
#include "../ppusbcomm_requests.h"
#include "hexdump.h"

int ppusbcomm_getConfig(raphnethid_hdl_t hdl, unsigned char param, unsigned char *rx, unsigned char rx_max)
{
	unsigned char cmd[2];
	int n;

	if (!hdl) {
		return -1;
	}

	cmd[0] = RQ_PPUSBCOMM_GET_CONFIG_PARAM;
	cmd[1] = param;

	n = raphnethid_exchange(hdl, cmd, 2, rx, rx_max);
	if (n<2)
		return n;

	n -= 2;

	// Drop the leading CMD and PARAM
	if (n) {
		memmove(rx, rx+2, n);
	}

	return n;
}
int ppusbcomm_setConfig(raphnethid_hdl_t hdl, unsigned char param, unsigned char *data, unsigned char len)
{
	unsigned char cmd[2 + len];
	int n;

	if (!hdl) {
		return -1;
	}

	cmd[0] = RQ_PPUSBCOMM_SET_CONFIG_PARAM;
	cmd[1] = param;
	memcpy(cmd + 2, data, len);

	n = raphnethid_exchange(hdl, cmd, 2 + len, cmd, sizeof(cmd));
	if (n<0)
		return n;

	return 0;
}

int ppusbcomm_getVersion(raphnethid_hdl_t hdl, char *dst, int dstmax)
{
	unsigned char cmd[32];
	int n;

	if (!hdl) {
		return -1;
	}

	if (dstmax <= 0)
		return -1;

	cmd[0] = RQ_PPUSBCOMM_GET_VERSION;

	n = raphnethid_exchange(hdl, cmd, 1, cmd, sizeof(cmd));
	if (n<0)
		return n;

	dst[0] = 0;
	if (n > 1) {
		strncpy(dst, (char*)cmd+1, n);
	}
	dst[dstmax-1] = 0;

	return 0;
}

int ppusbcomm_getSignature(raphnethid_hdl_t hdl, char *dst, int dstmax)
{
	unsigned char cmd[40];
	int n;

	if (!hdl) {
		return -1;
	}

	if (dstmax <= 0)
		return -1;

	cmd[0] = RQ_PPUSBCOMM_GET_SIGNATURE;

	n = raphnethid_exchange(hdl, cmd, 1, cmd, sizeof(cmd));
	if (n<0)
		return n;

	dst[0] = 0;
	if (n > 1) {
		strncpy(dst, (char*)cmd+1, n);
	}
	dst[dstmax-1] = 0;

	return 0;
}

int ppusbcomm_bootloader(raphnethid_hdl_t hdl)
{
	unsigned char cmd[4];
	int cmdlen;

	if (!hdl) {
		return -1;
	}

	cmd[0] = RQ_PPUSBCOMM_JUMP_TO_BOOTLOADER;
	cmdlen = 1;

	raphnethid_exchange(hdl, cmd, cmdlen, cmd, sizeof(cmd));

	return 0;
}

int ppusbcomm_uploadBlock(raphnethid_hdl_t hdl, unsigned char blk[32])
{
	unsigned char cmd[40];
	int cmdlen;

	if (!hdl) {
		return -1;
	}



	do {
		cmd[0] = RQ_PPUSBCOMM_UPLOAD_BLOCK;
		memcpy(cmd + 1, blk, 32);
		cmdlen = 33;

		if (raphnethid_exchange(hdl, cmd, cmdlen, cmd, sizeof(cmd)) < 0) {
			return -1;
		}
		printf("."); fflush(stdout);
		_delay_us(500);
	} while (cmd[1] == 1);

	return 0;
}

int ppusbcomm_uploadFile(raphnethid_hdl_t hdl, const char *filename, unsigned int flags)
{
	FILE *fptr;
	unsigned char blk[32];
	unsigned char transfer_header[32];
	int n;
	int b = 0;
	int filesize;
	unsigned short blocks;
	unsigned char extras;
	const char *target_filename;
	unsigned char lrc;

	if (!filename) {
		return -1;
	}

	target_filename = strrchr(filename, '/');
	if (!target_filename) {
		target_filename = filename;
	} else {
		// Skip the leading /
		target_filename++;
	}

	if (strlen(target_filename) > 12) {
		fprintf(stderr, "Filename too long. Please use DOS 8.3 format\n");
		return -1;
	}

	fptr = fopen(filename, "rb");
	if (!fptr) {
		perror("fopen");
		return -2;
	}

	fseek(fptr, 0, SEEK_END);
	filesize = ftell(fptr);
	fseek(fptr, 0, SEEK_SET);

	printf("File size: %d bytes\n", filesize);

	if (filesize/32 > 0xffff) {
		fprintf(stderr, "File too big for implementation.\n");
		fclose(fptr);
		return -2;
	}
	blocks = filesize/32;
	extras = filesize - (filesize/32)*32;
	printf("Blocks: %d\n", blocks);
	printf("Remainder bytes: %d\n", extras);

	/* First block: Transfer header*/

	/* 0-1: 'R','1'
	 * 2-14: Filename (8.3) zero-terminated
	 *
	 * 15-16: Number of 32-byte blocks
	 * 17: Extra bytes (remaining bytes that do not fill a block)
	 *
	 * 29-30 : File checksum (BSD style) TODO
	 * 31 : Header checksum (LRC)
	 *
	 * Note: Multi-byte values are little-endian
	 */
	memset(transfer_header, 0, sizeof(transfer_header));
	memcpy(transfer_header, "R1", 2); // Magic number and version
	strcpy((void*)transfer_header + 2, target_filename);
	transfer_header[15] = blocks & 0xff;
	transfer_header[16] = blocks >> 8;
	transfer_header[17] = extras;
	lrc = 0;
	for (n=0; n<sizeof(transfer_header)-1; n++) {
		lrc ^= transfer_header[n];
	}
	transfer_header[31] = lrc;

	if (flags & PPUSBCOMM_UPLOAD_FLAG_OMIT_HEADER) {
		printf("Not sending a header block\n");
	} else {
		printf("Sending header block\n");
		if (flags & PPUSBCOMM_UPLOAD_FLAG_DRY_RUN) {
			printHexBuf(transfer_header, 32);
		} else {
			if (ppusbcomm_uploadBlock(hdl, transfer_header)<0) {
				fclose(fptr);
				fprintf(stderr, "Error uploading header block\n");
				return -1;
			}
		}
	}

	printf("Sending file\n");
	do {
		memset(blk, 0, 32);
		n = fread(blk, 1, 32, fptr);
		if (n<0) {
			perror("fread");
			break;
		}
		if (n == 0) {
			break;
		}

//		printf("Uploading block %d\n", b);
		if (flags & PPUSBCOMM_UPLOAD_FLAG_DRY_RUN) {
			printHexBuf(blk, 32);
		} else {
			if (ppusbcomm_uploadBlock(hdl, blk)<0) {
				break;
			}
		}
		b++;

	} while (!feof(fptr));

	fclose(fptr);

	return 0;
}

